-----------------------------------------------------------------------------
-- Title      : SPI Slave Byte
-------------------------------------------------------------------------------
-- File       : spi_slave_byte.vhd
-- Author     : Grzegorz Daniluk
-- Company    : CERN
-------------------------------------------------------------------------------
--
-- Copyright (c) 2021 CERN / BE-CEM-EDL
--
-- This source file is free software; you can redistribute it   
-- and/or modify it under the terms of the GNU Lesser General   
-- Public License as published by the Free Software Foundation; 
-- either version 2.1 of the License, or (at your option) any   
-- later version.                                               
--
-- This source is distributed in the hope that it will be       
-- useful, but WITHOUT ANY WARRANTY; without even the implied   
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      
-- PURPOSE.  See the GNU Lesser General Public License for more 
-- details.                                                     
--
-- You should have received a copy of the GNU Lesser General    
-- Public License along with this source; if not, download it   
-- from http://www.gnu.org/licenses/lgpl-2.1.html
--
-------------------------------------------------------------------------------

-- CPOL = 0; CPHA = 0

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.gencores_pkg.all;

entity spi_slave_byte is
  generic (
    g_WSIZE : integer := 32);
  port (
    rst_n_i   : in std_logic;
    clk_sys_i : in std_logic;

    tx_dat_i     : in  std_logic_vector(g_WSIZE-1 downto 0);
    tx_load_i    : in  std_logic;
    rx_dat_o     : out std_logic_vector(g_WSIZE-1 downto 0);
    rx_ack_i     : in  std_logic;
    rx_newbyte_o : out std_logic;
    busy_o       : out std_logic;   -- clocking new byte
    ncs_o        : out std_logic;   -- copy of SPI nCS signal
    err_overrun_o  : out std_logic;
    err_underrun_o : out std_logic;

    spi_ncs_i  : in  std_logic;
    spi_sclk_i : in  std_logic;
    spi_mosi_i : in  std_logic;
    spi_miso_o : out std_logic
  );
end spi_slave_byte;

architecture behav of spi_slave_byte is

  signal sclk_posedge : std_logic;
  signal sclk_negedge : std_logic;
  signal ncs_synced   : std_logic;
  signal rx_newbyte     : std_logic;
  signal txing        : std_logic; -- '1' after first Tx word loaded
  signal new_txbyte   : std_logic;

  signal tx_data : std_logic_vector(g_WSIZE-1 downto 0);
  signal rx_data : std_logic_vector(g_WSIZE-1 downto 0);

  type t_state is (IDLE, TR_BYTE, PAUSE);
  signal state : t_state;
  signal cnt : unsigned(7 downto 0);

begin

  ncs_o <= ncs_synced;
  rx_dat_o <= rx_data;

  SCLK_SYNC: gc_sync_ffs
  generic map (
    g_SYNC_EDGE => "positive")
  port map (
    clk_i    => clk_sys_i,
    rst_n_i  => rst_n_i,
    data_i   => spi_sclk_i,
    npulse_o => sclk_negedge,
    ppulse_o => sclk_posedge);

  NCS_SYNC: gc_sync_ffs
  generic map (
    g_SYNC_EDGE => "positive")
  port map (
    clk_i    => clk_sys_i,
    rst_n_i  => rst_n_i,
    data_i   => spi_ncs_i,
    synced_o => ncs_synced);

  spi_miso_o <= tx_data(g_WSIZE-1);
  rx_newbyte_o <= rx_newbyte;

  process(clk_sys_i)
  begin
    if rising_edge(clk_sys_i) then
      if rst_n_i = '0' then
        cnt     <= (others=>'0');
        rx_data <= (others=>'0');
        tx_data <= (others=>'0');
        rx_newbyte <= '0';
        txing    <= '0';
        err_overrun_o  <= '0';
        err_underrun_o <= '0';
        busy_o   <= '0';
        state   <= IDLE;
      else

        ------------------------------
        -- FSM 
        ------------------------------
        case state is
          when IDLE =>
            cnt   <= (others=>'0');
            new_txbyte <= '0';
            err_overrun_o  <= '0';
            err_underrun_o <= '0';
            busy_o   <= '0';

            if rx_ack_i = '1' then
              rx_newbyte <= '0';
            end if;

            -- load tx word
            if tx_load_i = '1' then
              txing   <= '1';
              new_txbyte <= '1';
              tx_data <= tx_dat_i;
            else
              txing <= '0';
            end if;

            if ncs_synced = '0' then
              state <= PAUSE;
            end if;

          ----------------------------
          -- receive/transmit 1 byte
          when TR_BYTE =>
            busy_o   <= '1';

            if sclk_posedge = '1' then
              rx_data <= rx_data(g_WSIZE-2 downto 0) & spi_mosi_i;
            elsif sclk_negedge = '1' then
              tx_data <= tx_data(g_WSIZE-2 downto 0) & '0';
              cnt <= cnt + 1;
            end if;

            if ncs_synced = '1' then
              state <= IDLE;
            elsif cnt = g_WSIZE and rx_newbyte = '0' then
              new_txbyte <= '0';
              rx_newbyte <= '1';
              state <= PAUSE;
            elsif cnt = x"8" and rx_newbyte = '1' then
              new_txbyte <= '0';
              err_overrun_o <= '1';
              state <= PAUSE;
            end if;
              
          ----------------------------
          -- wait for next byte transfer or end of transaction
          when PAUSE =>
            cnt <= (others=>'0');
            busy_o   <= '0';

            if rx_ack_i = '1' then
              rx_newbyte <= '0';
            end if;

            -- load tx word
            if tx_load_i = '1' then
              txing   <= '1';
              new_txbyte <= '1';
              tx_data <= tx_dat_i;
            end if;

            if ncs_synced = '1' then
              state <= IDLE;
            elsif sclk_posedge = '1' then
              if (txing = '1' and new_txbyte = '0') then
                -- report underrun when we started providing Tx bytes and have
                -- not provided next TX word on time
                err_underrun_o <= '1';
              end if;
              -- begin of next byte transfer
              rx_data <= rx_data(g_WSIZE-2 downto 0) & spi_mosi_i;
              busy_o  <= '1';
              state   <= TR_BYTE;
            end if;

        end case;
        ------------------------------
        ------------------------------

      end if;
    end if;
  end process;

end behav;

