-----------------------------------------------------------------------------
-- Title      : SPI Bus Slave
-------------------------------------------------------------------------------
-- File       : spi_slave_test.vhd
-- Author     : Grzegorz Daniluk
-- Company    : CERN
-------------------------------------------------------------------------------
--
-- Copyright (c) 2021 CERN / BE-CEM-EDL
--
-- This source file is free software; you can redistribute it   
-- and/or modify it under the terms of the GNU Lesser General   
-- Public License as published by the Free Software Foundation; 
-- either version 2.1 of the License, or (at your option) any   
-- later version.                                               
--
-- This source is distributed in the hope that it will be       
-- useful, but WITHOUT ANY WARRANTY; without even the implied   
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      
-- PURPOSE.  See the GNU Lesser General Public License for more 
-- details.                                                     
--
-- You should have received a copy of the GNU Lesser General    
-- Public License along with this source; if not, download it   
-- from http://www.gnu.org/licenses/lgpl-2.1.html
--
-------------------------------------------------------------------------------
-- SPI commands | Cmd code (1B) | Adr Bytes | Data IN Bytes | Data Out Bytes | Comments
----------------+---------------+-----------+---------------+----------------+-
-- Dummy ID     | 0x00          | 0         | 0             | 4              | 0xCAFE<g_ID>
-- Write Data   | 0x02          | 3         | 1 to many     | 0              |
-- Read Data    | 0x03          | 3         | 0             | 1 to many      |
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- Address map:
-- 0x000 - 0x3FF - RAM
-- 0x400 -       - LEDs
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.genram_pkg.all;

entity spi_slave_test is
  generic (
    g_ID : integer := 0;
    g_LEDS_NUM  : integer := 6);
  port (
    rst_n_i   : in std_logic;
    clk_sys_i : in std_logic;

    spi_ncs_i  : in  std_logic;
    spi_sclk_i : in  std_logic;
    spi_mosi_i : in  std_logic;
    spi_miso_o : out std_logic;

    leds_o     : out std_logic_vector(g_LEDS_NUM-1 downto 0)
  );
end spi_slave_test;

architecture behav of spi_slave_test is

  constant c_POP_ID    : std_logic_vector(7 downto 0) := x"00";
  constant c_POP_RSTAT : std_logic_vector(7 downto 0) := x"01";
  constant c_POP_WRITE : std_logic_vector(7 downto 0) := x"02";
  constant c_POP_READ  : std_logic_vector(7 downto 0) := x"03";

  signal rxdata : std_logic_vector(31 downto 0);
  type t_state is (IDLE, CMD_DECODE,
    CMD_ID,
    CMD_WR, CMD_WR_MEM,
    MEM_RD, CMD_RD, CMD_RD_MEM,
    FIN);
  signal state : t_state;

  signal spi_rxdat, spi_txdat : std_logic_vector(31 downto 0);
  signal spi_rxword, spi_rxack : std_logic;
  signal spi_txload : std_logic;
  signal spi_ncs  : std_logic;

  -- CMD processing registers
  signal opadr : std_logic_vector(31 downto 0);
  alias r_op  is opadr(31 downto 24);
  alias r_adr is opadr(23 downto 0);

  -- Memory access signals
  signal mem_dat_in, mem_dat_out : std_logic_vector(31 downto 0);
  signal mem_adr : unsigned(31 downto 0);
  signal mem_wr  : std_logic;
  signal write   : std_logic;

  -- LEDs register
  signal leds_reg : std_logic_vector(31 downto 0);

begin

  U_SPI_BYTE: entity work.spi_slave_byte
  generic map (
    g_WSIZE => 32)
  port map (
    rst_n_i   => rst_n_i,
    clk_sys_i => clk_sys_i,
                   
    tx_dat_i     => spi_txdat,
    tx_load_i    => spi_txload,
    rx_dat_o     => spi_rxdat,
    rx_ack_i     => spi_rxack,
    rx_newbyte_o => spi_rxword,
    busy_o       => open,
    ncs_o        => spi_ncs,
    err_overrun_o  => open,
    err_underrun_o => open,
                   
    spi_ncs_i    => spi_ncs_i,
    spi_sclk_i   => spi_sclk_i,
    spi_mosi_i   => spi_mosi_i,
    spi_miso_o   => spi_miso_o);

  ------------------------------

  mem_wr <= write when(mem_adr < x"00000400") else
            '0';

  U_RAM: generic_spram
  generic map (
    g_data_width => 32,
    g_size => 1024,
    g_with_byte_enable => false)
  port map (
    rst_n_i => rst_n_i,
    clk_i   => clk_sys_i,
    bwe_i   => (others=>'1'),
    we_i    => mem_wr,
    a_i     => std_logic_vector(mem_adr(f_log2_size(1024)-1 downto 0)),
    d_i     => mem_dat_in,
    q_o     => mem_dat_out);

  ------------------------------

  process(clk_sys_i)
  begin
    if rising_edge(clk_sys_i) then
      if rst_n_i = '0' then
        spi_txload <= '0';
        spi_txdat  <= (others=>'0');
        spi_rxack  <= '0';
        opadr      <= (others=>'0');
        write      <= '0';
        mem_dat_in <= (others=>'0');
        mem_adr    <= (others=>'0');
        state <= IDLE;
      else
        ------------------------------
        -- FSM 
        ------------------------------
        case state is

          when IDLE =>
            spi_txload <= '0';
            spi_txdat  <= (others=>'0');
            spi_rxack  <= '0';
            write      <= '0';
            mem_dat_in <= (others=>'0');
            mem_adr    <= (others=>'0');

            if spi_ncs = '0' and spi_rxword = '1' then
              opadr     <= spi_rxdat;
              spi_rxack <= '1';
              state     <= CMD_DECODE;
            end if;

          when CMD_DECODE =>
            spi_txload <= '0';
            spi_txdat  <= (others=>'0');
            spi_rxack  <= '0';
            write      <= '0';
            mem_dat_in <= (others=>'0');
            mem_adr    <= (others=>'0');

            -- decode opcode
            if r_op = c_POP_ID then
              state <= CMD_ID;
            elsif r_op = c_POP_WRITE then
              -- adr from SPI is in words
              mem_adr <= x"00" & unsigned(r_adr);
              state <= CMD_WR;
            elsif r_op = c_POP_READ then
              -- adr from SPI is in words
              mem_adr <= x"00" & unsigned(r_adr);
              state <= MEM_RD;
            else
              state <= FIN;
            end if;

          ------------------------------
          -- OPCODE: ID
          ------------------------------
          when CMD_ID =>
            spi_txload <= '1';
            spi_txdat  <= x"CAFE" & std_logic_vector(to_unsigned(g_ID, 16));
            spi_rxack  <= '0';
            write      <= '0';
            mem_dat_in <= (others=>'0');
            mem_adr    <= (others=>'0');
            state <= FIN;

          ------------------------------
          -- OPCODE: WRITE
          ------------------------------
          when CMD_WR =>
            spi_txload <= '0';
            spi_txdat  <= (others=>'0');
            spi_rxack  <= '0';
            write      <= '0';
            mem_dat_in <= (others=>'0');
            if spi_ncs = '0' and spi_rxword = '1' then
              spi_rxack  <= '1';
              write      <= '1';
              mem_dat_in <= spi_rxdat;
              state <= CMD_WR_MEM;
            elsif spi_ncs = '1' then
              state <= FIN;
            end if;

          when CMD_WR_MEM =>
            spi_txload <= '0';
            spi_txdat  <= (others=>'0');
            spi_rxack  <= '0';
            write      <= '0';
            mem_adr    <= mem_adr + 1;  -- prepare addr for next write
            state      <= CMD_WR;

          ------------------------------
          -- OPCODE: READ
          ------------------------------
          when MEM_RD =>
            spi_txload <= '0';
            spi_txdat  <= (others=>'0');
            spi_rxack  <= '0';
            write      <= '0';
            mem_dat_in <= (others=>'0');
            state      <= CMD_RD;

          when CMD_RD =>
            spi_txload <= '1';
            spi_txdat  <= mem_dat_out;
            spi_rxack  <= '0';
            write      <= '0';
            mem_dat_in <= (others=>'0');
            mem_adr    <= mem_adr + 1;  -- prepare for next read
            state      <= CMD_RD_MEM;

          when CMD_RD_MEM =>
            spi_txload <= '0';
            spi_txdat  <= mem_dat_out;
            spi_rxack  <= '0';
            write      <= '0';
            mem_dat_in <= (others=>'0');
            if spi_ncs = '0' and spi_rxword = '1' then
              spi_rxack  <= '1';           -- discard received word
              state      <= CMD_RD;
            elsif spi_ncs = '1' then
              state <= FIN;
            end if;

          ------------------------------

          when FIN =>
          -- ack any Rx and just wait for the end of this transaction
            spi_txload <= '0';
            spi_txdat  <= (others=>'0');
            spi_rxack  <= '1';
            opadr      <= (others=>'0');
            write      <= '0';
            mem_dat_in <= (others=>'0');
            if spi_ncs = '1' then
              spi_txload <= '1';  -- write 0s as next tx word, just for cleanup
              state <= IDLE;
            end if;


        end case;
      end if;
    end if;
  end process;


  -----------------------------------------
  -- LEDs
  -----------------------------------------
  process(clk_sys_i)
  begin
    if rising_edge(clk_sys_i) then
      if rst_n_i = '0' then
        leds_reg <= (others=>'0');
      elsif write  = '1' and mem_adr = x"00000400" then
        leds_reg <= mem_dat_in;
      end if;
    end if;
  end process;

  leds_o <= leds_reg(g_LEDS_NUM-1 downto 0);


end behav;
