//-----------------------------------------------------------------------------
// Title      : Pentair FCM2 simulation interface
// Project    : Distributed I/O Tier
//-----------------------------------------------------------------------------
// File       : fcm2_interface.svh
// Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
// Company    : CERN (BE-CO-HT)
// Standard   : VHDL
//-----------------------------------------------------------------------------
// Description:
// This interface for FCM2 simulation model mimics Pentair FCM2 diagnostics
// module using i2c_slave from general-cores repository.
//-----------------------------------------------------------------------------
//
// Copyright (c) 2018 CERN
//
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 0.51 (the “License”) (which enables you, at your option,
// to treat this file as licensed under the Apache License 2.0); you may not
// use this file except in compliance with the License. You may obtain a copy
// of the License at http://solderpad.org/licenses/SHL-0.51.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//
//-----------------------------------------------------------------------------
//
// FCM2 commands layout:
// read i2c address: 0x39
// cmd 0: fan  status (return 1 Byte)
// cmd 1: temp status (return 1 Byte)
// cmd 2: volt status (return 1 Byte)
// cmd 3: fans speed  (return 12 Bytes) - 2B per fan
// cmd 4: temperature (return 6 Bytes)  - 1B per sensor
//
//-----------------------------------------------------------------------------

interface FCM_interface (input clk_i, rst_n_i);

  logic [6:0] dev_adr;
  logic ack;
  logic [7:0] tx_byte;
  logic [7:0] rx_byte;
  wire sta_p;
  wire sto_p;
  wire adr_good_p;
  wire r_done;
  wire w_done;

  reg [7:0] fan_status;
  reg [7:0] temp_status;
  reg [7:0] volt_status;
  reg [0:11] [7:0] fan_speed;
  reg [0:5] [7:0] temps;

  // set default values on reset
  initial begin
    dev_adr = 7'h1c;
    ack = 'b1;
    tx_byte = 'h00;
    fan_status  = 'hCA;
    temp_status = 'hFE;
    volt_status = 'hBE;
    fan_speed = '{'h01, 'h02, 'h03, 'h04, 'h05, 'h06, 'h07, 'h08, 'h09, 'h0A,
    'h0B, 'h0C};
    temps = '{'h11, 'h12, 'h13, 'h14, 'h15, 'h16};
  end

  task new_values;
  
    randomize (fan_status);
    randomize (temp_status);
    randomize (volt_status);
    randomize (fan_speed);
    randomize (temps);

  endtask : new_values

  // main functionality of FCM2, provide apropriate data depending on requested
  // command
  task reply;

    int cmd;

    $display("Execute reply");
    // react only on proper address
    wait(adr_good_p)
    @(posedge clk_i);
    // get command index
    wait(r_done)
    @(posedge clk_i);
    cmd = rx_byte;

    case(cmd)
      0: begin
        $display("Providing data for cmd0");
        tx_byte <= fan_status;
      end

      1: begin
        $display("Providing data for cmd1");
        tx_byte <= temp_status;
      end

      2: begin
        $display("Providing data for cmd2");
        tx_byte <= volt_status;
      end

      3: begin
        $display("Providing data for cmd3");
        for (int i=0; i<12; i++) begin
          tx_byte <= fan_speed[i];
          @(posedge w_done);
        end
        $display("Done");
      end

      4: begin
        $display("Providing data for cmd4");
        for (int i=0; i<6; i++) begin
          tx_byte <= temps[i];
          @(posedge w_done);
        end
        $display("Done");
      end

    endcase
    
  endtask : reply

endinterface
