///////////////////////////////////////////////////////////////////////////////
// Title      : Peripheral Board SPI Slave
// Project    : Distributed I/O Tier
///////////////////////////////////////////////////////////////////////////////
// File       : sim_main.sv
// Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
// Company    : CERN (BE-CO-HT)
// Standard   : SV
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2021 CERN
//
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 0.51 (the “License”) (which enables you, at your option,
// to treat this file as licensed under the Apache License 2.0); you may not
// use this file except in compliance with the License. You may obtain a copy
// of the License at http://solderpad.org/licenses/SHL-0.51.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////

`timescale 1ns/1ps
`include "if_wb_master.svh"
`include "simdrv_wb_spi.svh"
`include "simdrv_periph_spi.svh"
`include "sim_periph_test.svh"

module main;

  /////////////////////////////
  // clock and reset 
  logic clk_25m = 0;
  logic clk_25m_periph = 0;
  always #(20000ps) clk_25m <= ~clk_25m;

  // phase-shifted clk for SPI peripheral (slave)
  initial begin
    #10000ps;
    forever clk_25m_periph = #(20000ps) ~clk_25m_periph;
  end

  logic rst_n = 0;
  initial begin
    repeat(3) @(posedge clk_25m);
    rst_n = 1;
  end
  /////////////////////////////

  IWishboneMaster WB_spim (clk_25m, rst_n);
  logic spi_cs, spi_sclk, spi_mosi, spi_miso;
  logic leds[5];

  wb_spi #(
    .g_interface_mode(CLASSIC),
    .g_divider_len(16),
    .g_max_char_len(128),
    .g_num_slaves(1)
  ) SPIM (
    .clk_sys_i(clk_25m),
    .rst_n_i(rst_n),

    .wb_adr_i   (WB_spim.master.adr[4:0]),
    .wb_dat_i   (WB_spim.master.dat_o),
    .wb_dat_o   (WB_spim.master.dat_i),
    .wb_sel_i   (WB_spim.master.sel),
    .wb_stb_i   (WB_spim.master.stb),
    .wb_cyc_i   (WB_spim.master.cyc),
    .wb_we_i    (WB_spim.master.we),
    .wb_ack_o   (WB_spim.master.ack),
    .wb_err_o   (WB_spim.master.err),
    .wb_stall_o (WB_spim.master.stall),
    .int_o      (),

    .pad_cs_o   (spi_cs),
    .pad_sclk_o (spi_sclk),
    .pad_mosi_o (spi_mosi),
    .pad_miso_i (spi_miso)
  );
  assign WB_spim.master.rty = 1'b0;

  spi_slave_test #(
    .g_LEDS_NUM(5)
  ) PERIPH (
    .rst_n_i(rst_n),
    .clk_sys_i(clk_25m_periph),

    .spi_ncs_i(spi_cs),
    .spi_sclk_i(spi_sclk),
    .spi_mosi_i(spi_mosi),
    .spi_miso_o(spi_miso),

    .leds_o(leds)
  );


  /////////////////////////////
  // START
  /////////////////////////////
  initial begin
    uint32_t rval;
    uint32_t data[];
    uint32_t r_data[];
    CWishboneAccessor acc_spim;
    CSimDrv_WB_SPI spim_drv;
    CSimDrv_PERIPH_SPI periph;
    CSim_Periph_Test periph_test;

    // initialise OC SPI Master access
    acc_spim = WB_spim.get_accessor();
    acc_spim.set_mode(CLASSIC);
    WB_spim.settings.cyc_on_stall = 1;
    spim_drv = new(acc_spim, 0);

    #1us;
    // create Periph SPI object
    periph = new(spim_drv);
    // create Periph test object
    periph_test = new(periph);

    // Initialize and run some tests
    periph_test.init();
    periph_test.test_id();
    periph_test.read_write(10);
    //periph_test.leds(leds);

    $finish;

  end

endmodule
