-------------------------------------------------------------------------------
-- Title      : DIOT FMC DIO for RT Demonstrator
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : diot_periph_dio.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

entity diot_periph_dio is
  port (
    clk_25m_i : in std_logic;
    por_n_b   : inout std_logic;  -- Power-on-reset according to Application note AC380
    button_i  : in std_logic;

    -- FMC DIO
    fmc_p_i  : in  std_logic_vector(4 downto 0);
    fmc_n_i  : in  std_logic_vector(4 downto 0);
    fmc_p_o : out std_logic_vector(4 downto 0);
    fmc_n_o : out std_logic_vector(4 downto 0);
    fmc_term_en_o: out std_logic_vector(4 downto 0);
    fmc_oe_n     : out std_logic_vector(4 downto 0);

    fmc_led_top_o : out std_logic;
    fmc_led_bot_o : out std_logic;

    -- backplane connections
    slot_i   : in std_logic_vector(4 downto 0);
    slot_o   : out std_logic_vector(4 downto 0);
    slot_oen_i: in std_logic_vector(4 downto 0);

    leds_o  : out std_logic_vector(5 downto 0)
  );
end diot_periph_dio;

architecture rtl of diot_periph_dio is

  component BIBUF_LVCMOS25
  port (
    pad : inout std_logic;
    d   : in    std_logic;
    e   : in    std_logic;
    y   : out   std_logic);
  end component;

  component DFN1C1
  port (
    d   : in  std_logic;
    clk : in  std_logic;
    clr : in  std_logic;
    q   : out std_logic);
  end component;

  component INBUF_LVDS
  port (
    padp : in  std_logic;
    padn : in  std_logic;
    y    : out std_logic);
  end component;

  component OUTBUF_LVDS
  port (
    d    : in  std_logic;
    padp : out std_logic;
    padn : out std_logic);
  end component;

  signal por_n_buf : std_logic;
  signal por_n_d   : std_logic_vector(1 downto 0);
  signal cnt       : unsigned(24 downto 0);
  signal leds      : std_logic_vector(5 downto 0);
  signal rst_cnt   : unsigned(15 downto 0) := (others=> '0');
  signal rst_n     : std_logic;

  signal fmc_input : std_logic_vector(4 downto 0);
  signal fmc_output: std_logic_vector(4 downto 0);
begin

-------------------------------------------------------------------------------
-- Power-on-reset according to Microsemi Application Note AC380
-------------------------------------------------------------------------------
  cmp_por_bibuf: BIBUF_LVCMOS25
  port map (
    pad => por_n_b,
    d   => '0',
    e   => '1',
    y   => por_n_buf);

  cmd_por_DFF1: DFN1C1
  port map (
    d   => '1',
    clk => clk_25m_i,
    clr => por_n_buf,
    q   => por_n_d(0));

  cmd_por_DFF2: DFN1C1
  port map (
    d   => por_n_d(0),
    clk => clk_25m_i,
    clr => por_n_buf,
    q   => por_n_d(1));

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if (por_n_d(1) = '0' or button_i = '1') then
        rst_cnt <= (others=>'0');
      elsif and_reduce(std_logic_vector(rst_cnt)) /= '1' then
        rst_cnt <= rst_cnt + 1;
      end if;
    end if;
  end process;
  rst_n <= '1' when and_reduce(std_logic_vector(rst_cnt))='1' else
           '0';

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
  fmc_term_en_o <= (others=>'0');
  fmc_oe_n      <= slot_oen_i;

  GEN_INBUFS_I: for I in 0 to 4 generate
    U_BUF_FMC: INBUF_LVDS
    port map (
      padp => fmc_p_i(I),
      padn => fmc_n_i(I),
      y    => fmc_input(I));

    slot_o <= fmc_input;
  end generate; 

  GEN_OUTBUFS_I: for I in 0 to 4 generate
    fmc_output <= slot_i;
    U_BUF_BCKPL: OUTBUF_LVDS
    port map (
      d    => fmc_output(I),
      padp => fmc_p_o(I),
      padn => fmc_n_o(I));
  end generate; 

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if rst_n = '0' then
        cnt  <= (others=>'0');
        leds <= "000001";
      else
        if cnt = to_unsigned(25000000, 25) then
          cnt <= (others=>'0');
          leds(5 downto 0) <= leds(4 downto 0) & leds(5);
        else
          cnt <= cnt + 1;
        end if;
      end if;
    end if;
  end process;

  fmc_led_top_o <= leds(0);
  fmc_led_bot_o <= slot_i(3);
  GEN_LEDS_O: for I in 0 to 5 generate
    leds_o(I) <= '0' when leds(I) = '1' else
                 'Z';
  end generate;

-------------------------------------------------------------------------------

end rtl;
