-------------------------------------------------------------------------------
-- Title      : DIOT WIC Demonstrator
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : diot_wic_top.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- DIOT top module with for Warm Interlocks Controller demo.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the License) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- AS IS BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;
use work.genram_pkg.all;
library proasic3e;
use proasic3e.all;

entity diot_urv_top is
  port (
    clk_25m_i : in std_logic;
    por_n_b   : inout std_logic;  -- Power-on-reset according to Application note AC380
    button_i  : in std_logic;

    -- nanoFIP Wishbone
    dat_i  : in  std_logic_vector(7 downto 0);
    dat_o  : out std_logic_vector(7 downto 0);
    adr_o  : out std_logic_vector(9 downto 0);
    stb_o  : out std_logic;
    wclk_o : out std_logic;
    we_o   : out std_logic;
    cyc_o  : out std_logic;
    ack_i  : in  std_logic;
    rst_o  : out std_logic;     --  Active high
    -- nanoFIP control
    rstin_o    : out std_logic; --  Active low
    nostat_o   : out std_logic;
    var3_acc_o : out std_logic;
    var3_rdy_i : in  std_logic;
    var2_acc_o : out std_logic;
    var2_rdy_i : in  std_logic;
    var1_acc_o : out std_logic;
    var1_rdy_i : in  std_logic;
    p3_lgth_o  : out std_logic_vector(2 downto 0);

    -- backplane connections
    -- in the first HW demonstrator not all slots are fully populated
    -- crate monitoring
    diag_scl_b  : inout std_logic;
    diag_sda_b  : inout std_logic;
--
--    shared_bus : inout std_logic_vector(3 downto 0);
--    irq_n_i    : in  std_logic;

    s1_p_b  : inout std_logic_vector(15 downto 0);
    s1_n_b  : inout std_logic_vector(15 downto 0);
    s1_en_i : in    std_logic;
    s2_p_b  : inout std_logic_vector(13 downto 0); 
    s2_n_b  : inout std_logic_vector(13 downto 0); 
    s2_en_i : in    std_logic;
    s3_p_b  : inout std_logic_vector(13 downto 0);
    s3_n_b  : inout std_logic_vector(13 downto 0);
    s3_en_i : in    std_logic;
    s4_p_b  : inout std_logic_vector(13 downto 0);
    s4_n_b  : inout std_logic_vector(13 downto 0);
    s4_en_i : in    std_logic;
    s5_p_b  : inout std_logic_vector(13 downto 0);
    s5_n_b  : inout std_logic_vector(13 downto 0);
    s5_en_i : in    std_logic;
    -- slot6 not available in first HW demonstrator
    s7_p_b  : inout std_logic_vector(15 downto 0);
    s7_n_b  : inout std_logic_vector(15 downto 0);
    s7_en_i : in    std_logic;
    -- slot8 not available in first HW demonstrator

    leds_o  : out std_logic_vector(5 downto 0)
  );
end diot_urv_top;

architecture rtl of diot_urv_top is

  component BIBUF_LVCMOS25
  port (
    pad : inout std_logic;
    d   : in    std_logic;
    e   : in    std_logic;
    y   : out   std_logic);
  end component;

  component INBUF
  port (
    pad : in    std_logic;
    y   : out   std_logic);
  end component;

  component PLLINT
  port (
    a   : in    std_logic;
    y   : out   std_logic);
  end component;

  component DFN1C1
  port (
    d   : in  std_logic;
    clk : in  std_logic;
    clr : in  std_logic;
    q   : out std_logic);
  end component;

  constant c_SLOT_NUM     : integer := 6;
  constant c_PERIOD       : unsigned(24 downto 0) := to_unsigned(25000000, 25);

  signal por_n_buf : std_logic;
  signal por_n_d   : std_logic_vector(1 downto 0);
  signal leds      : std_logic_vector(5 downto 0);
  signal rst_n     : std_logic;
  signal arst      : std_logic;
  
  signal plc_rst_n : std_logic;

  type t_slot_vec is array (c_SLOT_NUM-1 downto 0) of std_logic_vector (31 downto 0);

  signal slots_loops        : t_slot_vec;
  signal slots_relays       : t_slot_vec;
  signal var1_rdy_synced : std_logic;
  signal var3_rdy_synced : std_logic;

  signal wb_fip_in : t_wishbone_slave_in;
  signal wb_fip_out : t_wishbone_slave_out;
  signal sel : std_logic_vector(3 downto 0);
  signal unc_dat : std_logic_vector(31 downto 8);
  signal unc_adr : std_logic_vector(31 downto 0);

  signal var1_acc, var1_rdy, var1_read : std_logic;
  signal var3_acc, var3_rdy : std_logic;

  --  IO bus for the plc cpu
  signal wb_plc_in : t_wishbone_slave_in;
  signal wb_plc_out : t_wishbone_slave_out;

  signal plc_sram_addr : std_logic_vector(12 downto 2);
  signal plc_sram_dato : std_logic_vector(31 downto 0);
  signal plc_sram_dati : std_logic_vector(31 downto 0);
  signal plc_sram_wr   : std_logic;

  --  MEM bus for FIP cpu to write into PLC memory.
  signal wb_mem_in : t_wishbone_slave_in;
  signal wb_mem_out : t_wishbone_slave_out;

  --  Mailboxes (from the FIP cpu POV)
  signal mbox_out_reg : std_logic_vector(31 downto 0);
  signal mbox_out_dat : std_logic_vector(31 downto 0);
  signal mbox_out_wr : std_logic;
  signal mbox_out_rd : std_logic;
  signal mbox_out_rdy : std_logic;
  signal mbox_in_reg : std_logic_vector(31 downto 0);
  signal mbox_in_dat : std_logic_vector(31 downto 0);
  signal mbox_in_wr : std_logic;
  signal mbox_in_rd : std_logic;
  signal mbox_in_rdy : std_logic;

  signal fip_cyc : std_logic;
  signal fip_stb : std_logic;
  signal fip_we : std_logic;
  signal fip_ack : std_logic;
  signal fip_adr : std_logic_vector (9 downto 0);
  signal fip_dato : std_logic_vector (7 downto 0);
  signal fip_dati : std_logic_vector (7 downto 0);

  signal clk_25m_in : std_logic;
  signal clk : std_logic;
  signal pll_lock : std_logic;
  signal clk_2 : std_logic;
begin
  inst_clk_pad : PLLINT
    port map (
      a => clk_25m_i,
      y => clk_25m_in
    );

  inst_pll: entity work.pll_25_50
    port map (
      powerdown => '1',
      clka => clk_25m_in,
      lock => pll_lock,
      gla => clk,
      glb => clk_2
  );

-------------------------------------------------------------------------------
-- Power-on-reset according to Microsemi Application Note AC380
-------------------------------------------------------------------------------
  cmp_por_bibuf: BIBUF_LVCMOS25
  port map (
    pad => por_n_b,
    d   => '0',
    e   => '1',
    y   => por_n_buf);

  cmd_por_DFF1: DFN1C1
  port map (
    d   => '1',
    clk => clk,
    clr => por_n_buf,
    q   => por_n_d(0));

  cmd_por_DFF2: DFN1C1
  port map (
    d   => por_n_d(0),
    clk => clk,
    clr => por_n_buf,
    q   => por_n_d(1));

  arst <= not por_n_d(1) or not pll_lock or button_i;

  inst_reset: entity work.gc_reset_multi_aasd
    generic map (
      g_CLOCKS => 1,
      g_RST_LEN => 16
    )
    port map (
      arst_i => arst,
      clks_i (0) => clk,
      rst_n_o (0) => rst_n
    );

-------------------------------------------------------------------------------
-- Synchronizers for RDY signals (40MHz clock domain in nanoFIP
-------------------------------------------------------------------------------
  cmp_sync_ffs_var1: entity work.gc_sync
    port map(
      clk_i      => clk,
      rst_n_a_i  => rst_n,
      d_i        => var1_rdy_i,
      q_o        => var1_rdy_synced);
  cmp_sync_ffs_var3: entity work.gc_sync
    port map(
      clk_i    => clk,
      rst_n_a_i  => rst_n,
      d_i   => var3_rdy_i,
      q_o => var3_rdy_synced);

  process (clk, rst_n)
  begin
    if rising_edge(clk) then
      if rst_n = '0' then
        --  Variable was not read.
        var1_read <= '0';
      else
        if var1_rdy_synced = '0' then
          --  A new value will be written.
          var1_read <= '0';
        elsif var1_acc = '1' then
          --  Variable is (being) read.
          var1_read <= '1';
        end if;
      end if;
    end if;
  end process;

  --  A variable is ready if it the data are ready from nanofip and not yet read.
  var1_rdy <= var1_rdy_synced and not var1_read;

  var3_rdy <= var3_rdy_synced;

  var1_acc_o <= var1_acc;
  var3_acc_o <= var3_acc;

--  Mailboxes (from FIP POV).
  process(clk)
  begin
    if rising_edge(clk) then
      if rst_n = '0' then
        mbox_in_rdy <= '0';
        mbox_out_rdy <= '0';
        mbox_out_reg <= (others => '0');
        mbox_in_reg <= x"11_22_33_44";
      else
        if mbox_out_wr = '1' then
          mbox_out_rdy <= '1';
          mbox_out_reg <= mbox_out_dat;
        elsif mbox_out_rd = '1' then
          mbox_out_rdy <= '0';
        end if;

        if mbox_in_wr = '1' then
          mbox_in_rdy <= '1';
          mbox_in_reg <= mbox_in_dat;
        elsif mbox_in_rd = '1' then
          mbox_in_rdy <= '0';
        end if;
      end if;
    end if;
  end process;

-------------------------------------------------------------------------------

  inst_fip_urv: entity work.fip_urv
    generic map (
      g_IRAM_LOG_SIZE => 10,
      g_DRAM_LOG_SIZE => 11,
      g_iram_init => "../../../sw/fip_urv/fip_urv.ram"
    )
    port map (
      clk_sys_i => clk,
      rst_n_i => rst_n,
      dwb_o => wb_fip_in,
      dwb_i => wb_fip_out
    );

  inst_fip_regs: entity work.fip_urv_regs
    port map (
      rst_n_i => rst_n,
      clk_i => clk,
      wb_i => wb_fip_in,
      wb_o => wb_fip_out,
      plc_ctrl_rstn_o => plc_rst_n,
      fip_status_var1_rdy_i => var1_rdy,
      fip_status_var3_rdy_i => var3_rdy,
      fip_var1_acc_o => var1_acc,
      fip_var3_acc_o => var3_acc,
      mailboxes_mboxout_o  => mbox_out_dat,
      mailboxes_mboxout_wr_o => mbox_out_wr,
      mailboxes_mboxin_i   => mbox_in_reg,
      mailboxes_mboxin_rd_o => mbox_in_rd,
      mailboxes_status_mbin_i => mbox_in_rdy,
      mailboxes_status_mbout_i => mbox_out_rdy,
      presence_en_i (0) => s1_en_i,
      presence_en_i (1) => s2_en_i,
      presence_en_i (2) => s3_en_i,
      presence_en_i (3) => s4_en_i,
      presence_en_i (4) => s5_en_i,
      presence_en_i (5) => s7_en_i,
      presence_en_i (6) => '0',
      presence_en_i (7) => '0',
      leds_val_o => leds,
      boards_pins_0_i (15 downto 0) => slots_loops(0)(15 downto 0),
      boards_pins_0_i (31 downto 16) => slots_relays(0)(15 downto 0),
      boards_pins_1_i (15 downto 0) => slots_loops(1)(15 downto 0),
      boards_pins_1_i (31 downto 16) => slots_relays(1)(15 downto 0),
      boards_pins_2_i (15 downto 0) => slots_loops(2)(15 downto 0),
      boards_pins_2_i (31 downto 16) => slots_relays(2)(15 downto 0),
      boards_pins_3_i (15 downto 0) => slots_loops(3)(15 downto 0),
      boards_pins_3_i (31 downto 16) => slots_relays(3)(15 downto 0),
      boards_pins_4_i (15 downto 0) => slots_loops(4)(15 downto 0),
      boards_pins_4_i (31 downto 16) => slots_relays(4)(15 downto 0),
      boards_pins_5_i (15 downto 0) => slots_loops(5)(15 downto 0),
      boards_pins_5_i (31 downto 16) => slots_relays(5)(15 downto 0),
      boards_pins_6_i => x"00_00_00_00",
      boards_pins_7_i => x"00_00_00_00",
      fip_reg_i.dat (7 downto 0) => fip_dati,
      fip_reg_i.dat (31 downto 8) => x"00_00_00",
      fip_reg_i.ack => fip_ack,
      fip_reg_i.err => '0',
      fip_reg_i.rty => '0',
      fip_reg_i.stall => '0',
      fip_reg_o.dat (7 downto 0) => fip_dato,
      fip_reg_o.dat (31 downto 8) => unc_dat,
      fip_reg_o.adr (1 downto 0) => unc_adr (1 downto 0),
      fip_reg_o.adr (11 downto 2) => fip_adr,
      fip_reg_o.adr (31 downto 12) => unc_adr (31 downto 12),
      fip_reg_o.cyc => fip_cyc,
      fip_reg_o.stb => fip_stb,
      fip_reg_o.sel => sel,
      fip_reg_o.we => fip_we,
      plc_mem_i => wb_mem_out,
      plc_mem_o => wb_mem_in
    );

  --  Simple speed adapter for WB with nanofip.
  process (clk_2)
  begin
    if rising_edge(clk_2) then
      if rst_n = '0' then
        cyc_o <= '0';
        stb_o <= '0';
        we_o <= '0';
      else
        cyc_o <= fip_cyc;
        stb_o <= fip_stb;
        we_o  <= fip_we;
        dat_o <= fip_dato;
        adr_o <= fip_adr;
        fip_dati <= dat_i;
      end if;
    end if;
  end process;

  inst_fip_ack: entity work.gc_sync
  port map (
    clk_i => clk_2,
    rst_n_a_i => rst_n,
    d_i => ack_i,
    q_o => fip_ack
  );

  wclk_o <= clk_2;

  rst_o <= not rst_n;
  nostat_o <= '1';
  rstin_o <= rst_n;

  var2_acc_o <= '0';
  p3_lgth_o  <= "001"; -- 8 bytes

  diag_scl_b <= 'Z';
  diag_sda_b <= 'Z';

-------------------------------------------------------------------------------

  inst_plc_urv: entity work.plc_urv
    generic map (
      g_iram_log_size => 10
    )
    port map (
      clk_sys_i => clk,
      rst_n_i => plc_rst_n,
      ram_wb_i => wb_mem_in,
      ram_wb_o => wb_mem_out,
      dwb_o => wb_plc_in,
      dwb_i => wb_plc_out
    );

  inst_plc_regs: entity work.plc_urv_regs
  port map (
    rst_n_i => plc_rst_n,
    clk_i => clk,
    wb_i => wb_plc_in,
    wb_o => wb_plc_out,
    sram_addr_o => plc_sram_addr,
    sram_data_i => plc_sram_dati,
    sram_data_o => plc_sram_dato,
    sram_wr_o   => plc_sram_wr,
    mailboxes_mboxout_o  => mbox_in_dat,
    mailboxes_mboxout_wr_o => mbox_in_wr,
    mailboxes_mboxin_i   => mbox_out_reg,
    mailboxes_mboxin_rd_o => mbox_out_rd,
    mailboxes_status_mbin_i => mbox_out_rdy,
    mailboxes_status_mbout_i => mbox_in_rdy,
    presence_en_i (0) => s1_en_i,
    presence_en_i (1) => s2_en_i,
    presence_en_i (2) => s3_en_i,
    presence_en_i (3) => s4_en_i,
    presence_en_i (4) => s5_en_i,
    presence_en_i (5) => s7_en_i,
    presence_en_i (6) => '0',
    presence_en_i (7) => '0',
    loops_pins_0_i => slots_loops(0),
    loops_pins_1_i => slots_loops(1),
    loops_pins_2_i => slots_loops(2),
    loops_pins_3_i => slots_loops(3),
    loops_pins_4_i => slots_loops(4),
    loops_pins_5_i => slots_loops(5),
    loops_pins_6_i => x"0000_0000",
    loops_pins_7_i => x"0000_0000",
    relays_pins_0_o => slots_relays(0),
    relays_pins_1_o => slots_relays(1),
    relays_pins_2_o => slots_relays(2),
    relays_pins_3_o => slots_relays(3),
    relays_pins_4_o => slots_relays(4),
    relays_pins_5_o => slots_relays(5),
    relays_pins_6_o => open,
    relays_pins_7_o => open
  );

  p_ram: process (clk)
  is
    constant c_DRAM_LOG_SIZE : natural := 12;
    variable mem : t_ram32_type (2**(c_DRAM_LOG_SIZE - 2) - 1 downto 0);
    variable addr : natural range mem'range;
  begin
    if rising_edge(clk) then
      addr := to_integer(unsigned(plc_sram_addr(c_DRAM_LOG_SIZE - 1 downto 2)));
      plc_sram_dati <= mem(addr);
      if plc_sram_wr = '1' then
        for i in 0 to 3 loop
          if true then -- if reg_dm_data_select (i) = '1' then
            mem(addr)(8*i + 7 downto 8*i) := plc_sram_dato(8*i + 7 downto 8*i);
          end if;
        end loop;
      end if;
    end if;
  end process;

-------------------------------------------------------------------------------
  GEN_REAL_IO: block
  begin
    -- slots assignment
    gen_slots_0: for i in 15 downto 0 generate
      inst_sync: entity work.gc_sync
        port map (
          clk_i => clk,
          rst_n_a_i => rst_n,
          d_i => s1_p_b(i),
          q_o => slots_loops(0)(i)
        );
    end generate;
    slots_loops(0)(31 downto 16) <= (others=>'0');

    gen_slots_1: for i in 13 downto 0 generate
      inst_sync: entity work.gc_sync
        port map (
          clk_i => clk,
          rst_n_a_i => rst_n,
          d_i => s2_p_b(i),
          q_o => slots_loops(1)(i)
        );
    end generate;
    slots_loops(1)(31 downto 14) <= (others=>'0');

    gen_slots_2: for i in 13 downto 0 generate
      inst_sync: entity work.gc_sync
        port map (
          clk_i => clk,
          rst_n_a_i => rst_n,
          d_i => s3_p_b(i),
          q_o => slots_loops(2)(i)
        );
    end generate;
    slots_loops(2)(31 downto 14) <= (others=>'0');

    gen_slots_3: for i in 13 downto 6 generate
      inst_sync: entity work.gc_sync
        port map (
          clk_i => clk,
          rst_n_a_i => rst_n,
          d_i => s4_p_b(i),
          q_o => slots_loops(3)(i)
        );
    end generate;
    inst_slots_3_3: entity work.gc_sync
       port map (
         clk_i => clk,
         rst_n_a_i => rst_n,
         d_i => s4_p_b(3),
         q_o => slots_loops(3)(3)
       );
    slots_loops(3)(31 downto 14) <= (others=>'0');
    slots_loops(3)(5 downto 4) <= (others=>'0');
    slots_loops(3)(2 downto 0) <= (others=>'0');

    gen_slots_4: for i in 13 downto 0 generate
      inst_sync: entity work.gc_sync
        port map (
          clk_i => clk,
          rst_n_a_i => rst_n,
          d_i => s5_p_b(i),
          q_o => slots_loops(4)(i)
        );
    end generate;
    slots_loops(4)(31 downto 14) <= (others=>'0');

    gen_slots_5: for i in 15 downto 0 generate
      inst_sync: entity work.gc_sync
        port map (
          clk_i => clk,
          rst_n_a_i => rst_n,
          d_i => s7_p_b(i),
          q_o => slots_loops(5)(i)
        );
    end generate;
    slots_loops(5)(31 downto 16) <= (others=>'0');

    s1_n_b(15 downto 0) <= slots_relays(0)(15 downto 0);
    s2_n_b(13 downto 0) <= slots_relays(1)(13 downto 0);
    s3_n_b(13 downto 0) <= slots_relays(2)(13 downto 0);
    s4_n_b(13 downto 6) <= slots_relays(3)(13 downto 6);
    s4_n_b(3)           <= slots_relays(3)(3);
    s5_n_b(13 downto 0) <= slots_relays(4)(13 downto 0);
    s7_n_b(15 downto 0) <= slots_relays(5)(15 downto 0);
  end block;

  --  Leds: assign to '0' to switch on the led, 'Z' to switch off.
  leds_o(0) <= '0' when rst_n = '0' or leds(0) = '1' else 'Z';
  leds_o(1) <= '0' when pll_lock = '0' or leds(1) = '1' else 'Z';

  GEN_LEDS_O: for I in 2 to 5 generate
    leds_o(I) <= '0' when leds(I) = '1' else
                 'Z';
  end generate;
end rtl;
