-------------------------------------------------------------------------------
-- Title      : DIOT System Board v1 (GEFE-based) as Peripheral with SPI
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : sb_periph_spi.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CEM-EDL)
-- Standard   : VHDL
-------------------------------------------------------------------------------
--
-- Copyright (c) 2021 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

entity sb_periph_spi is
  port (
    clk_25m_i : in std_logic;
    por_n_b   : inout std_logic;  -- Power-on-reset according to Application note AC380

    -- backplane connections
    spi_ncs_p_i  : in  std_logic_vector(3 downto 0);
    spi_ncs_n_i  : in  std_logic_vector(3 downto 0);
    spi_sclk_p_i : in  std_logic_vector(3 downto 0);
    spi_sclk_n_i : in  std_logic_vector(3 downto 0);
    spi_mosi_p_i : in  std_logic_vector(3 downto 0);
    spi_mosi_n_i : in  std_logic_vector(3 downto 0);
    spi_miso_p_o : out std_logic_vector(3 downto 0);
    spi_miso_n_o : out std_logic_vector(3 downto 0);

    leds_o  : out std_logic_vector(5 downto 0)
  );
end sb_periph_spi;

architecture rtl of sb_periph_spi is
  constant c_NUM_SPI : integer := 4;

  component INBUF_LVDS
  port (
    padp : in  std_logic;
    padn : in  std_logic;
    y    : out std_logic);
  end component;

  component OUTBUF_LVDS
  port (
    d    : in  std_logic;
    padp : out std_logic;
    padn : out std_logic);
  end component;

  component BIBUF_LVCMOS25
  port (
    pad : inout std_logic;
    d   : in    std_logic;
    e   : in    std_logic;
    y   : out   std_logic);
  end component;

  component DFN1C1
  port (
    d   : in  std_logic;
    clk : in  std_logic;
    clr : in  std_logic;
    q   : out std_logic);
  end component;

  signal por_n_buf : std_logic;
  signal por_n_d   : std_logic_vector(1 downto 0);
  signal rst_cnt   : unsigned(15 downto 0) := (others=> '0');
  signal rst_n     : std_logic;


  signal spi_ncs  : std_logic_vector(3 downto 0);
  signal spi_sclk : std_logic_vector(3 downto 0);
  signal spi_mosi : std_logic_vector(3 downto 0);
  signal spi_miso : std_logic_vector(3 downto 0);
  signal leds     : std_logic_vector(5 downto 0);

begin

-------------------------------------------------------------------------------
-- Power-on-reset according to Microsemi Application Note AC380
-------------------------------------------------------------------------------
  cmp_por_bibuf: BIBUF_LVCMOS25
  port map (
    pad => por_n_b,
    d   => '0',
    e   => '1',
    y   => por_n_buf);

  cmd_por_DFF1: DFN1C1
  port map (
    d   => '1',
    clk => clk_25m_i,
    clr => por_n_buf,
    q   => por_n_d(0));

  cmd_por_DFF2: DFN1C1
  port map (
    d   => por_n_d(0),
    clk => clk_25m_i,
    clr => por_n_buf,
    q   => por_n_d(1));

  process(clk_25m_i)
  begin
    if rising_edge(clk_25m_i) then
      if (por_n_d(1) = '0') then
        rst_cnt <= (others=>'0');
      elsif and_reduce(std_logic_vector(rst_cnt)) /= '1' then
        rst_cnt <= rst_cnt + 1;
      end if;
    end if;
  end process;
  rst_n <= '1' when and_reduce(std_logic_vector(rst_cnt))='1' else
           '0';
-------------------------------------------------------------------------------

    SPI0: entity work.spi_slave_test
    generic map (
      g_ID => 0,
      g_LEDS_NUM => 6)
    port map (
      rst_n_i     => rst_n,
      clk_sys_i   => clk_25m_i,
                
      spi_ncs_i   => spi_ncs(0),
      spi_sclk_i  => spi_sclk(0),
      spi_mosi_i  => spi_mosi(0),
      spi_miso_o  => spi_miso(0),
                
      leds_o      => leds);

  GEN_SPI: if c_NUM_SPI > 1 generate
    SPI_CHANNELS: for I in 1 to c_NUM_SPI-1 generate
      SPI: entity work.spi_slave_test
      generic map (
        g_ID => I,
        g_LEDS_NUM => 6)
      port map (
        rst_n_i     => rst_n,
        clk_sys_i   => clk_25m_i,
                  
        spi_ncs_i   => spi_ncs(I),
        spi_sclk_i  => spi_sclk(I),
        spi_mosi_i  => spi_mosi(I),
        spi_miso_o  => spi_miso(I),
                  
        leds_o      => open);
    end generate;
  end generate;

  -- assign not used MISO outputs
  GEN_NO_SPI: if c_NUM_SPI < 4 generate
    NO_SPI: for I in c_NUM_SPI to 3 generate
      spi_miso(I) <= '0';
    end generate;
  end generate;

  GEN_LEDS: for I in 0 to 5 generate
    leds_o(I) <= '0' when leds(I) = '1' else
                 'Z';
  end generate;

-------------------------------------------------------------------------------
-- I/O LVDS buffers
-------------------------------------------------------------------------------
  GEN_SPI_BUFS: for I in 0 to 3 generate
    U_NCS: INBUF_LVDS
    port map (
      padp => spi_ncs_p_i(I),
      padn => spi_ncs_n_i(I),
      y    => spi_ncs(I));

    U_SCLK: INBUF_LVDS
    port map (
      padp => spi_sclk_p_i(I),
      padn => spi_sclk_n_i(I),
      y    => spi_sclk(I));

    U_MOSI: INBUF_LVDS
    port map (
      padp => spi_mosi_p_i(I),
      padn => spi_mosi_n_i(I),
      y    => spi_mosi(I));

    U_MISO: OUTBUF_LVDS
    port map (
      d    => spi_miso(I),
      padp => spi_miso_p_o(I),
      padn => spi_miso_n_o(I));
  end generate;

end rtl;
