/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstring>

#include "nanofip.h"

using namespace std::placeholders;

nanoFIP::nanoFIP(unsigned int addr, unsigned int variable_size)
    : AgentBase(addr), prod_dirty(true)
{
    cons_data.resize(variable_size);
    prod_data.resize(variable_size);
    ClearDataStats();
}


int nanoFIP::AddVariables(WindowPeriodicVar& window)
{
    int err;

    // Produced variable
    err = window.AddVariable( (PROD_VAR_ID << 8) | address,
            MSTRFIP_DATA_FLAGS_PROD, GetVarSize(),
            std::bind( &nanoFIP::prodVarHandler, this, _1, _2, _3 ) );

    if (err)
        return err;

    // Consumed variable
    err = window.AddVariable( (CONS_VAR_ID << 8) | address,
            MSTRFIP_DATA_FLAGS_CONS, GetVarSize(),
            std::bind( &nanoFIP::consVarHandler, this, _1, _2, _3 ) );

    return err;
}


void nanoFIP::ClearDataStats()
{
    const mstrfip_data_status types[] = {
        MSTRFIP_DATA_OK, MSTRFIP_DATA_FRAME_ERROR,
        MSTRFIP_DATA_PAYLOAD_ERROR, MSTRFIP_DATA_NOT_RECEIVED
    };

    for(auto type : types) {
        data_stats[type] = 0;
    }
}


const std::string& nanoFIP::GetDataStatDesc(mstrfip_data_status stat)
{
    static const std::unordered_map<mstrfip_data_status, std::string> statusMap = {
        { MSTRFIP_DATA_OK,              "Data OK" },
        { MSTRFIP_DATA_FRAME_ERROR,     "Frame error" },
        { MSTRFIP_DATA_PAYLOAD_ERROR,   "Payload error" },
        { MSTRFIP_DATA_NOT_RECEIVED,    "Data not received" }
    };

    return statusMap.at(stat);
}


void nanoFIP::consVarHandler(mstrfip_dev *dev, mstrfip_data *cvar,
    mstrfip_irq* irq)
{
    {
        std::lock_guard<std::mutex> lock(cons_mtx);
        mstrfip_var_update(dev, cvar);

        switch (cvar->status) {
        case MSTRFIP_DATA_OK:
            memcpy((void*) cons_data.data(), cvar->buffer, cons_data.size());
            break;

        case MSTRFIP_DATA_PAYLOAD_ERROR:
            if (cvar->payload_error & MSTRFIP_FRAME_PAYLOAD_NOT_REFRESH)
                reportError(NOT_FRESHED);
            if (cvar->payload_error & MSTRFIP_FRAME_PAYLOAD_NOT_SIGNIFICANT)
                reportError(NOT_SIGNIFICANT);
            DBG("Data payload error\n");
            break;

        case MSTRFIP_DATA_NOT_RECEIVED:
            reportError(MISSED);
            DBG("Data not received\n");
            break;

        case MSTRFIP_DATA_FRAME_ERROR:
            if (cvar->frame_error & MSTRFIP_FRAME_TMO)
                reportError(TIMEOUT);
            if (cvar->frame_error & MSTRFIP_FRAME_ERR)
                reportError(HW_ERR);
            if (cvar->frame_error & MSTRFIP_FRAME_BAD_CTRL)
                reportError(HW_CTRL);
            if (cvar->frame_error & MSTRFIP_FRAME_BAD_PDU)
                reportError(HW_PDU);
            if (cvar->frame_error & MSTRFIP_FRAME_BAD_BSZ)
                reportError(HW_BSZ);
            DBG("Data frame error\n");
            break;

        default:
            reportError(UNKNOWN);
            DBG("Unknown error\n");
            break;
        }

        ++data_stats[cvar->status];
    }

    if (cons_cb)        // User callback
        cons_cb(*this, cvar);
}


void nanoFIP::prodVarHandler(mstrfip_dev *dev, mstrfip_data *pvar,
    mstrfip_irq* irq)
{
    {
        std::lock_guard<std::mutex> lock(prod_mtx);

        if (prod_dirty) {
            memcpy((void*) pvar->buffer, (void*) prod_data.data(), prod_data.size());
            int res = mstrfip_var_write(dev, pvar);

            if (res) {
                reportError(WRITE_FAILED);
                DBG("Variable write failed\n");
            }
            else
            {
                prod_dirty = false;
            }
        }
    }

    if (prod_cb)        // User callback
        prod_cb(*this, pvar);
}
